import { DominoesGame } from "./DominoesGame";
import { PlayerAI } from "./entities/player/PlayerAI";
import { PlayerInterface } from "./entities/player/PlayerInterface";
import { LoggingService } from "../common/LoggingService";  
import { getRandomSeed, uuid, wait } from "../common/utilities";
import { GameSessionState } from "./dto/GameSessionState";
import { PlayerNotificationManager } from './PlayerNotificationManager';
import seedrandom, { PRNG } from "seedrandom";

export class GameSession {
  private game: DominoesGame | null = null;
  private minHumanPlayers: number = 1;  
  private waitingForPlayers: boolean = true;
  private waitingSeconds: number = 0;
  private logger: LoggingService = new LoggingService();
  private mode: string = 'classic';
  private pointsToWin: number = 100;
  private playerNotificationManager: PlayerNotificationManager;
  id: string;
  players: PlayerInterface[] = [];
  sessionInProgress: boolean = false;
  maxPlayers: number = 4;
  seed!: string
  rng!: PRNG

  constructor(public creator: PlayerInterface, public name?: string) {
    this.playerNotificationManager = new PlayerNotificationManager(this);
    this.id = uuid();
    this.name = name || `Game ${this.id}`;
    this.addPlayer(creator);
    this.logger.info(`GameSession created by: ${creator.name}`);
    this.creator = creator;
    this.playerNotificationManager.notifySessionState();
  }
  
  get numPlayers() {
    return this.players.length;
  }

  private async startGame(seed: string) {
    this.rng = seedrandom(seed);
    const missingPlayers = this.maxPlayers - this.numPlayers;
    for (let i = 0; i < missingPlayers; i++) {
      this.addPlayer(this.createPlayerAI(i));
    }
    this.game = new DominoesGame(this.players, this.rng);
    this.sessionInProgress = true;
    this.logger.info('Game started');
    this.playerNotificationManager.notifySessionState();
    await this.game.start();
    return this.endGame();
  }

  private endGame(): any {
    if (this.game !== null) {
      this.sessionInProgress = false;
      const { gameBlocked, gameTied, winner } = this.game;
      
      gameBlocked ? console.log('Game blocked!') : gameTied ? console.log('Game tied!') : console.log('Game over!');
      console.log('Winner: ' + winner?.name + ' with ' + winner?.pipsCount() + ' points');
      this.getScore(this.game);
      this.sessionInProgress = false;
      this.logger.info('Game ended');
      this.game = null;
      this.playerNotificationManager.notifySessionState();

      return {
        gameBlocked,
        gameTied,
        winner
      };
    }
  }

  private getScore(game: DominoesGame) {
    const pips = game.players
      .sort((a,b) => (b.pipsCount() - a.pipsCount()))
      .map(player => {
        return `${player.name}: ${player.pipsCount()}`;
      });
    console.log(`Pips count: ${pips.join(', ')}`);
    const totalPoints = game.players.reduce((acc, player) => acc + player.pipsCount(), 0);
    if (game.winner !== null) {
      game.winner.score += totalPoints;
    }
    const scores = game.players
      .sort((a,b) => (b.score - a.score))
      .map(player => {
        return `${player.name}: ${player.score}`;
      });
    console.log(`Scores: ${scores.join(', ')}`);
  }

  createPlayerAI(i: number) {
    const AInames = ["Alice (AI)", "Bob (AI)", "Charlie (AI)", "David (AI)"];
    return new PlayerAI(AInames[i], this.rng);
  }

  async start(seed?: string) {
    this.seed = seed || getRandomSeed();
    console.log('seed :>> ', this.seed);
    if (this.sessionInProgress) {
      throw new Error("Game already in progress");
    }
    this.waitingForPlayers = true;
    this.logger.info('Waiting for players to join');
    while (this.numPlayers < this.maxPlayers) {
      this.waitingSeconds += 1;
      this.logger.info(`Waiting for players to join: ${this.waitingSeconds}`);
      await wait(1000);
    }
    this.waitingForPlayers = false;
    this.logger.info('All players joined');
    this.startGame(this.seed);
  }
  
  addPlayer(player: PlayerInterface) {
    if (this.numPlayers >= this.maxPlayers) {
      throw new Error("GameSession is full");
    }
    this.players.push(player);
    this.logger.info(`${player.name} joined the game!`);
  }

  toString() {
    return `GameSession:(${this.id} ${this.name})`;
  }

  getState(): GameSessionState {
    return {
      id: this.id,
      name: this.name!,
      creator: this.creator.id,
      players: this.players.map(player =>( {
        id: player.id,
        name: player.name,
      })),
      sessionInProgress: this.sessionInProgress,
      maxPlayers: this.maxPlayers,
      numPlayers: this.numPlayers,
      waitingForPlayers: this.waitingForPlayers,
      waitingSeconds: this.waitingSeconds,
      seed: this.seed,
      mode: this.mode,
      pointsToWin: this.pointsToWin,
      status: this.sessionInProgress ? 'in progress' : 'waiting'
    };
  }
}